import os, random

WHITESPACES = [' ', '\r', '\t', '\n']


def rndstr(l):
    s = ''
    for i in range(l):
        s += chr(random.randint(97, 122))
    return s


def gen_name(dir, ext, l=8):
    s = rndstr(l) + ext
    while os.path.exists(os.path.join(dir, s)):
        s = rndstr(l) + ext
    return s


def working_dir():
    return os.getcwd()


def str_equal(str, segments, idx):
    l = len(segments)
    if idx is None:
        return False
    if l == 1:
        str_len = len(segments[0])
        return str[idx:idx+str_len] == segments[0]
    else:
        str_len = len(segments[0])
        if str[idx:idx+str_len] != segments[0]:
            return False
        return str_equal(str, segments[1:], cross_whitespaces(str, idx+str_len-1))


def count(str, k, syntax=False, exp = False):
    '''
    syntax & exp ARGS CANNOT BOTH BE TRUE!!!
    OTHERWISE PROGRAM WILL BE UNPREDICTABLE
    '''
    if exp:
        segments = k.split('*')
    else:
        segments = k

    r = 0
    l = len(k)

    if str_equal(str, segments, 0):
        if syntax:
            if not str[l].isalnum():
                r += 1
        else:
            r += 1

    for i in range(1, len(str)):
        if str_equal(str, segments, i):
            if syntax:
                if not str[i - 1].isalnum() and not str[i + l].isalnum():
                    r += 1
            else:
                r += 1

    return r


def skip_whitespaces(str, index):
    i = index + 1
    if i == len(str):
        return None
    while str[i] in WHITESPACES:
        i += 1
        if i == len(str):
            return None

    return str[i]


def cross_whitespaces(str, index):
    i = index + 1
    if i == len(str):
        return None
    while str[i] in WHITESPACES:
        i += 1
        if i == len(str):
            return None

    return i


def count_fors(str):
    r = 0
    for i in range(len(str) - 3):
        if (str[i:i+3] == 'for') and (not str[i-1].isalnum()) and (skip_whitespaces(str, i+2) == '('):
            r += 1

    return r


def min_semis(proc):
    fors = count_fors(proc)
    if fors > 0:
        ret = fors * 2
    elif '#' in proc or count(proc, 'template', syntax=True) > 0:
        ret = 0
    else:
        ret = 1
    return ret


def isblock(proc):
    ret = True

    if count(proc, '{') > count(proc, '}'):
        ret = False

    if count(proc, ';') < min_semis(proc):
        ret = False

    return ret


def empty_couts(str):
    proc, dictionary = replace_string(str)
    ret = ''
    in_cout = False

    for i in range(len(proc)):
        if not in_cout and i < len(proc)-4:
            if (proc[i:i+4] == 'cout') and (not proc[i-1].isalnum()) and (skip_whitespaces(proc, i+3) == '<'):
                in_cout = True
                ret += 'cout << "";'
            else:
                ret += proc[i]

        elif in_cout and proc[i] == ';':
            in_cout = False

        elif not in_cout:
            ret += proc[i]

    for i in dictionary:
        ret = ret.replace(i, dictionary[i])

    return ret


def replace_string(s, str_tok_pre='~STR', str_tok_suf='~'):
    ret = ''
    count = 0
    dictionary = {}
    tok = None

    in_string = False
    in_char = False

    for i in range(len(s)):
        if s[i] == '"' and s[i - 1] != '\\' and not in_char:
            in_string = not in_string
            if in_string:
                tok = str_tok_pre + str(count) + str_tok_suf
                ret += tok
                dictionary[tok] = '"'
                count += 1
            else:
                dictionary[tok] += '"'

        elif s[i] == "'" and s[i - 1] != '\\' and not in_string:
            in_char = not in_char
            if in_char:
                tok = str_tok_pre + str(count) + str_tok_suf
                ret += tok
                dictionary[tok] = "'"
                count += 1
            else:
                dictionary[tok] += "'"

        elif not in_char and not in_string:
            ret += s[i]

        else:
            dictionary[tok] += s[i]

    return ret, dictionary


def replace_comment(s, com_tok_pre='~COM', com_tok_suf='~'):
    ret = ''
    count = 0
    dictionary = {}
    tok = None

    in_single = False
    in_multi = False

    for i in range(len(s)):
        if (s[i] == '/' and s[i - 1] == '/') and (not in_single and not in_multi):
            in_single = True
            tok = com_tok_pre + str(count) + com_tok_suf
            ret = ret[:-1] + tok
            dictionary[tok] = '//'
            count += 1

        elif (s[i] == '\n' and s[i - 1] != '\\') and in_single:
            in_single = False
            ret += '\n'
            dictionary[tok] += '\n'

        elif (s[i] == '*' and s[i - 1] == '/') and (not in_single and not in_multi):
            in_multi = True
            tok = com_tok_pre + str(count) + com_tok_suf
            ret = ret[:-1] + tok
            dictionary[tok] = '/*'
            count += 1

        elif (s[i] == '/' and s[i - 1] == '*') and in_multi:
            in_multi = False
            dictionary[tok] += '/'

        elif not in_single and not in_multi:
            ret += s[i]

        else:
            dictionary[tok] += s[i]

    return ret, dictionary



import os


def build(filedir, filename, outname):
    cmd = ['cd {}'.format(filedir),
           'g++ -o {} {}'.format(outname, filename)]
    cmd = ' & '.join(cmd)

    a = os.system(cmd)

    if a != 0:
        print('Compilation terminated with return code {}'.format(a))
        return False
    else:
        return True


def run(filedir, filename):
    cmd = ['cd {}'.format(filedir),
           '{}'.format(filename)]
    cmd = ' & '.join(cmd)

    return os.system(cmd)



from scierra.utils.utils import rndstr
from scierra.utils.utils import gen_name
from scierra.utils.utils import working_dir
from scierra.utils.utils import str_equal
from scierra.utils.utils import count
from scierra.utils.utils import skip_whitespaces
from scierra.utils.utils import cross_whitespaces
from scierra.utils.utils import count_fors
from scierra.utils.utils import min_semis
from scierra.utils.utils import isblock
from scierra.utils.utils import empty_couts
from scierra.utils.utils import replace_string
from scierra.utils.utils import replace_comment



from scierra.sim import Simulator

def main():
    flag = True
    res = True
    simul = Simulator()

    print('Scierra C++ Interpreter\nCopyright (c) PerceptronV 2020')
    line = input('\n++> ')
    stripline = line.strip()

    if stripline == '<print>':
        simul.print_code()
    elif stripline == '<restart>':
        simul.restart()
    elif stripline == '<esc>':
        flag = False
    elif stripline == '':
        pass
    else:
        res = simul.addline(line + '\n')

    while (flag):
        if res:
            line = input('\n++> ')
        else:
            line = input('\n--> ')
        stripline = line.strip()

        if stripline == '<print>':
            simul.print_code()
        elif stripline == '<restart>':
            res = True
            simul.restart()
        elif stripline == '<esc>':
            flag = False
        elif stripline == '':
            pass
        else:
            res = simul.addline(line + '\n')



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scierra",
    version="0.5",
    author="PerceptronV",
    author_email="neutrinovs@gmail.com",
    description="A Simulated C++ Interpreter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PerceptronV/scierra",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts':[
            'scierra = scierra.scierra:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Interpreters",
        "Development Status :: 4 - Beta",
        "Natural Language :: English"
    ],
    python_requires='>=3.3',
)


from scierra.compile import build, run
from scierra.utils import gen_name, working_dir, count, replace_string, replace_comment, empty_couts, isblock
import os
import platform

KEYWORDS = {
    'exp_prep':[
        '#*include',
        '#*define'
    ],
    'prep': [
        'typedef',
        'using'
    ],
    'glob': [
        'class',
        'struct',
        'return',
        'void',
        'template',
        'typename'
    ]
}

FILE_EXTENSION = {
    'Windows': '.exe',
    'Linux': '.out',
    'Darwin': '.out'
}


class Simulator(object):
    def __init__(self):
        self.wdir = working_dir()
        self.preprocs = ""
        self.globals = ""
        self.mains = ""
        self.buff = ""
        self.TEMPLATE = "#include<iostream>\n#include<sstream>\n#include<fstream>\n#include<vector>\n#include<string>\nusing namespace std;\n{}\n{}\n{}\n"

    def gen_code(self, preps=None, globs=None, mains=None):
        if preps is None:
            preps = self.preprocs
        if globs is None:
            globs = self.globals
        if mains is None:
            mains = self.mains

        code = self.TEMPLATE.format(
            preps, globs, 'int main(){\n' + mains + '\nreturn 0;\n}'
        )

        return code

    def addline(self, line):
        self.buff += line
        proc, str_dict = replace_string(self.buff)
        proc, com_dict = replace_comment(proc)

        if isblock(proc):
            section = 'main'

            if '<prep>' in proc:
                section = 'prep'
                proc = proc.replace('<prep>', '')
            elif '<glob>' in proc:
                section = 'glob'
                proc = proc.replace('<glob>', '')
            elif '<main>' in proc:
                section = 'main'
                proc = proc.replace('<main>', '')
            else:
                for i in KEYWORDS['glob']:
                    if count(proc, i, syntax=True) > 0:
                        section = 'glob'

                for i in KEYWORDS['prep']:
                    if count(proc, i, syntax=True) > 0:
                        section = 'prep'

                for i in KEYWORDS['exp_prep']:
                    if count(proc, i, exp=True) > 0:
                        section = 'prep'

            for i in str_dict:
                proc = proc.replace(i, str_dict[i])
            for i in com_dict:
                proc = proc.replace(i, com_dict[i])
            self.buff = proc

            if section == 'prep':
                if self.invoke(self.preprocs + self.buff, self.globals, empty_couts(self.mains)):
                    self.preprocs += self.buff
            elif section == 'glob':
                if self.invoke(self.preprocs, self.globals + self.buff, empty_couts(self.mains)):
                    self.globals += self.buff
            elif section == 'main':
                if self.invoke(self.preprocs, self.globals, empty_couts(self.mains) + self.buff):
                    self.mains += self.buff

            self.buff = ""
            return True

        elif proc[0] == '<' and proc[:6] not in ['<prep>', '<glob>', '<main>']:
            self.buff = 'cout << ' + self.buff[1:] + ';'
            self.invoke(self.preprocs, self.globals, empty_couts(self.mains) + self.buff)
            self.buff = ""
            return True

        else:
            return False

    def invoke(self, preps, globs, mains, src_name=None, out_name=None):
        code = self.gen_code(preps, globs, mains)

        if src_name is None:
            src_name = gen_name(self.wdir, '.cpp')
        if out_name is None:
            out_name = gen_name(self.wdir, FILE_EXTENSION[platform.system()])

        f = open(src_name, 'w')
        f.write(code)
        f.close()

        ret = build(self.wdir, src_name, out_name)
        if ret:
            ret = run(self.wdir, out_name) == 0
            os.remove(os.path.join(self.wdir, out_name))

        os.remove(os.path.join(self.wdir, src_name))

        return ret

    def print_code(self):
        print(self.gen_code())

    def restart(self):
        self.__init__()
